/*
 * libkarma/rio_rw.c
 *
 * Copyright (c) Frank Zschockelt <libkarma@freakysoft.de> 2004
 * 
 * You may distribute and modify this program under the terms of 
 * the GNU GPL, version 2 or later.
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include <taglib/tag_c.h>

#include "mp3.h"
#include "wav.h"

#include "karma.h"
#include "properties.h"
#include "util.h"
#include "errors.h"
#include "playlist.h"
#include "rio_rw.h"
#include "utf8.h"

/* #define BLOCKSIZE 32768 */
#define BLOCKSIZE 131072

void set_tag_props(HASH *p, const char *filename);
void add_tag(HASH *p, char *key, char *value);

int synchronize_necessary(int rio)
{
	int ret;
	char *device_generation;
	char *p;

	device_generation = 
               strdup(lk_properties_get_property(0, "device_generation"));
	lk_karma_get_device_settings(rio, &p);
	lk_properties_import(p);
	free(p);
	ret=strcmp(device_generation, 
                   lk_properties_get_property(0, "device_generation"));
	free(device_generation);
	return ret;
}

int rio_read_playlist(int rio, uint32_t fid, const char * filename)
{
        playlist *pl;
        int dst;

        pl=lk_playlist_fid_read(fid);
        if (pl==NULL)
	         return -1;  /* Can't get playlist for this fid */
	
	if (mk_path(filename))
           fprintf(stderr,
                   "** Could not create directory of path: %s\n", filename);
        dst=open(filename,
                 O_CREAT|O_EXCL|O_WRONLY,S_IREAD|S_IWRITE|S_IRGRP|S_IROTH);
        if(dst==-1)
                 return -1;  /* Can't open: e.g., file already exist */

        write(dst, pl->playlist, pl->length);
        close(dst); free(pl->playlist); free(pl);

        return 0;
}

int rio_read(int rio, uint32_t fid, const char * filename)
{
	int dst;
	uint64_t blk=BLOCKSIZE, offs=0, filesize=0;
	uint64_t got=0;
	char *tmp, *tmp1;

/* 	dst=creat(filename, O_EXCL); */
	if (mk_path(filename))
           fprintf(stderr,
                   "** Could not create directory of path: %s\n", filename);
	dst=open(filename,
		 O_CREAT|O_EXCL|O_WRONLY,S_IREAD|S_IWRITE|S_IRGRP|S_IROTH);
	if(dst==-1) 
	        return -1; /* Can not open (e.g., file already exist) */

	tmp1=lk_properties_get_property(fid, "length");
	if(tmp1 != NULL)
	        filesize=atoll(tmp1);
	else
	        return -1; /* tmp1==NULL => filesize undefined */
        if(filesize<=0)
	        return -1; /* Mystery: filesize sometimes can be 0 ???!!! */

	while(offs<filesize){
		if(lk_karma_read_file_chunk(rio,offs,blk,fid,&tmp,&got) != 0){
			close(dst);
			unlink(filename);
			return -1;
		}
		write(dst, tmp, got);
		free(tmp);
		offs+=got;
		if((filesize-offs) < blk)
			blk=filesize-offs;
	}
	
	close(dst);
	return 0;
}

/*
	 To use that function properly:
	 	1. Get an IO Lock for Writing
	 	2. call synchronize_necessary 
	 	   and synchronize the properties if necessary ;)
	 	3. use rio_write as often as you need to :)
	 	4. release the IO Lock.
*/
uint32_t rio_write(int rio, const char * filename)
{
	int got=-1,fd;
	uint32_t new_fid;
	uint64_t offs=0;
	const uint64_t siz=BLOCKSIZE;
	char * p, tmp[BLOCKSIZE];
	struct stat * len;
	HASH * props;
	
	len=malloc(sizeof(struct stat));
	if(lstat(filename, len)==-1)
		return -1;
	
	new_fid=lk_properties_new_property(); 
        /* uses lk_properties_getnextfid()...*/
	props=lk_properties_idsearch(new_fid);
	if(props==NULL){
		printf("huh, no hash found?\n");
		return -1;
	}
	p=strrchr(filename, '.');
	
	add_general(props, filename,len->st_size);
	if(p != NULL){
		++p;	
		if(strcmp(p, "ogg") == 0)
			got=get_ogg_props(props, filename, len->st_size);
		else if(strcmp(p, "flac") == 0)
			got=get_flac_props(props, filename, len->st_size);
		else if(strcmp(p, "mp3") == 0)
			got=get_mp3_props(props, filename);
		else if(strcmp(p, "wav") == 0)
			got=get_wave_props(props,filename);
		/*
			 else if(strcmp(p, "wma") == 0)
			 got=get_wma_props(props,filename);
		 */	
		else 
			got=get_taxi_props(props);
	}else got=get_taxi_props(props);

	free(len);
	
	if(got==0){
		fd=open(filename, O_RDONLY);
		while((got=read(fd, &tmp, siz)) > 0){
			/*if(got < siz) 
				memset(&tmp[got], 0, (siz-got)%4);
			if(lk_karma_write_file_chunk(
			     rio, offs, got+((siz-got)%4), new_fid, 0, tmp)!=0)
			    */
			if(lk_karma_write_file_chunk(
			     rio, offs, got, new_fid, 0, tmp)!=0)
			{
			       printf("lk_karma_write_file_chunk: %u of %llu\n",
				                                      got, siz);
			       lk_properties_del_property(new_fid);
			       return -1;
			}
			offs+=got;
		}
		close(fd);
		
		p=lk_properties_export(new_fid);
		got=lk_karma_update_file_details(rio, new_fid, p);
		free(p);
/* 		return got; */
		return new_fid;
	}else{
		lk_properties_del_property(new_fid);
		return -1;
	}
}

void add_general(HASH * p, const char * file, unsigned int len)
{
	time_t generation_time;
	char *t;
	
	generation_time=time(NULL);
	lk_properties_set_property_hash(p, "fid_generation", 
                                           simple_itoa(generation_time));
	lk_properties_set_property_hash(p, "ctime", 
                                           simple_itoa(generation_time));
	lk_properties_set_property_hash(p, "length", simple_itoa(len));
	t=strrchr(file, '/');
	t=utf8_to_codeset((char *)(t?t+1:file));
	lk_properties_set_property_hash(p, "title", t);
	free(t);
}

int get_ogg_props(HASH *p, const char * oggfile, unsigned int len)
{
	int rid_fd;
	
	lk_properties_set_property_hash(p, "codec", "vorbis");
	lk_properties_set_property_hash(p, "type", "tune");
	
	rid_fd=open(oggfile, O_RDONLY);
	lk_properties_set_property_hash(
	      p, "rid", 
              (char*)lk_generate_rid(rid_fd, 0, len));
	close(rid_fd);
	
	set_tag_props(p, oggfile);
	return 0;
}

int get_flac_props(HASH *p, const char * flacfile, unsigned int len)
{
	int rid_fd;

	/*default values*/
	lk_properties_set_property_hash(p, "source", NOT_TAGGED);
	lk_properties_set_property_hash(p, "artist", NOT_TAGGED);
	
	lk_properties_set_property_hash(p, "codec", "flac");
	lk_properties_set_property_hash(p, "type", "tune");
	
	rid_fd=open(flacfile, O_RDONLY);
	lk_properties_set_property_hash(p, "rid", 
                          (char*)lk_generate_rid(rid_fd, 0, len));
	close(rid_fd);
	
	set_tag_props(p, flacfile);
	return 0;
}

void add_tag(HASH *p, char *key, char *value)
{
  if (value[0])
     lk_properties_set_property_hash(p, key, value);

  /*printf("Key = %s ; Value = %s\n", key, value);*/
}

/*** You can use this with all tag-formats, which taglib supports.
  */
void set_tag_props(HASH *p, const char *filename)
{
	TagLib_File *tl_file;
	TagLib_Tag *tl_tag;
	const TagLib_AudioProperties * tl_audio;
	int i;
	char bitrate[6];

  	tl_file = taglib_file_new(filename);
  	tl_tag = taglib_file_tag(tl_file);
  	tl_audio = taglib_file_audioproperties(tl_file);

  
	add_tag(p, "title", taglib_tag_title(tl_tag));
	add_tag(p, "artist", taglib_tag_artist(tl_tag));
	add_tag(p, "source", taglib_tag_album(tl_tag));
	add_tag(p, "comment", taglib_tag_comment(tl_tag));
	add_tag(p, "genre", taglib_tag_genre(tl_tag));

	i = taglib_tag_year(tl_tag);
	if (i)
		add_tag(p, "year", simple_itoa(i)); 
	i = taglib_tag_track(tl_tag);
	if (i) {
		add_tag(p, "tracknr", simple_itoa(i));
		add_tag(p, "file_id", simple_itoa(i));
	}
	i=taglib_audioproperties_samplerate(tl_audio);
	if(i)
		add_tag(p, "samplerate", simple_itoa(i));
	i=taglib_audioproperties_length(tl_audio)*1000;
	if(i)
		add_tag(p, "duration", simple_itoa(i));
	strcpy(bitrate, "vs");
	i=taglib_audioproperties_bitrate(tl_audio);
	if(i){
		if(i>999) strcat(bitrate, "999");
		else strcat(bitrate, simple_itoa(i));
		add_tag(p, "bitrate", bitrate);
	}

	taglib_file_free(tl_file);
 	/*taglib_tag_free_strings();*/
}

int get_mp3_props(HASH *p, const char * mp3file)
{
	mp3info mp3;
	char bitrate[5];
	int rid_fd;
	
	lk_properties_set_property_hash(p, "type", "tune");
	lk_properties_set_property_hash(p, "codec", "mp3");
	/*default values*/
	lk_properties_set_property_hash(p, "artist", NOT_TAGGED);
	lk_properties_set_property_hash(p, "source", NOT_TAGGED);
	
	memset(&mp3,0,sizeof(mp3info));

	mp3.filename=(char *)mp3file;
	mp3.file=fopen(mp3file, "r");
	if(mp3.file==NULL) return -1;
	get_mp3_info(&mp3);
	fclose(mp3.file);
	
	lk_properties_set_property_hash(p, "offset", simple_itoa(mp3.offset));
	rid_fd=open(mp3file, O_RDONLY);
	lk_properties_set_property_hash(p, "rid", 
	   (char *)lk_generate_rid(rid_fd, mp3.offset, mp3.datasize));
	close(rid_fd);
        
        set_tag_props(p, mp3file);
	
	if(!mp3.vbr){
		strcpy(bitrate, "fs");
		strcat(bitrate, simple_itoa(header_bitrate(&mp3.header)));
		lk_properties_set_property_hash(p, "bitrate", bitrate);
	}
	
        return 0;
}

int get_taxi_props(HASH *p)
{
	lk_properties_set_property_hash(p, "duration", "0");
	lk_properties_set_property_hash(p, "offset", "0");
	lk_properties_set_property_hash(p, "bitrate", "fs128");
	lk_properties_set_property_hash(p, "codec", "taxi");
	lk_properties_set_property_hash(p, "type", "taxi");
	
	return 0;
}
/*
int get_wma_props(HASH *p, const char * wmafile)
{
	return 0;
}
*/
int get_wave_props(HASH *p, const char * wavefile)
{
	int rid_fd;
	wave_header *wh;
	
	if(openwav(&wh, (char*)wavefile) == -1){
		cleanup(&wh);
		return 1;
	}
	lk_properties_set_property_hash(p, "codec", "wave");
	lk_properties_set_property_hash(p, "type", "tune");
	lk_properties_set_property_hash(p, "bitrate", "fs999");
	lk_properties_set_property_hash(p, "source", "wave_file");
	lk_properties_set_property_hash(p, "artist", "wave_file");
	lk_properties_set_property_hash(p, "offset", 
                                           simple_itoa(sizeof(wave_header)));
	lk_properties_set_property_hash(p, "samplerate", 
                                           simple_itoa(wh->nSamplesPerSec));
	lk_properties_set_property_hash(p, "duration", simple_itoa(1000 * 
	      wh->RiffSize/wh->nChannels/(wh->nChannels * wh->nSamplesPerSec)));
	lk_properties_set_property_hash(p, "file_id", "0");
	lk_properties_set_property_hash(p, "tracknr", "0");
	
	rid_fd=open(wavefile, O_RDONLY);
	lk_properties_set_property_hash(p, "rid", 
   		    (char*)lk_generate_rid(rid_fd, 0, wh->RiffSize/1024));
	close(rid_fd);

	cleanup(&wh);	
	return 0;
}
