/*
 * libkarma/riocp.c
 *
 * Copyright (c) Frank Zschockelt <libkarma@freakysoft.de> 2004-2005
 *               Enrique Vidal 2004-2005
 *
 * You may distribute and modify this program under the terms of 
 * the GNU GPL, version 2 or later.
 *
 */
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#ifdef LINUX
#include <ftw.h>
#else
#include "ftw.h"
#endif
#include <locale.h>
#include <langinfo.h>

#include "karma.h"
#include "ssdp.h"
#include "properties.h"
#include "util.h"
#include "errors.h"
#include "fdb.h"
#include "rio_rw.h"
#include "utf8.h"
#include "pathedit.h"

#define CHECK(x) (x)?"failed":"ok"

#define defaultCodeset "ISO8859-15"

#define ssdpRetry 3                     /* number of retries allowed for SSDP */

void usage(void);
void currentKarma(int i, int *k);
int treecopy(const char *file, const struct stat *sb, int flag);
uint32_t * searchStr(int exact, char *searchString);
double getUsedSize(int karma);

void fidNamesFileHandler(const int op, const int fid, const char * data);
void deleteFids(uint32_t *vp);
int isodigit (char d);
char esc(char c);
char *expand(char *str);
void downloadFids(uint32_t *vp, int doDownload,
		char* useTuneProps, char* useTaxiProps, char* usePlstProps,
		char* edPathStr1, char* edPathStr2, char* edPathStr3,
		int squeezeRep);


void usage(void)
{
  printf("usage: riocp [-a {path|hostname|ip}] [-p password] [-c charset]\n"
         "             [-f [fidNameFile]] [-s searchString ]\n"
	 "             [-u props] [-x props] [-l props] [-e set1 set2 set3\n" 
         "             [-q] [-d] [-r] [-R] [files_to_upload ...]\n"
	 "\n");
  printf("  \"-f\" download fids with filenames given in fidNamesFile,\n"
	 "       where fidNamesFile defaults to ~/.openrio/.../fidNamesFile\n"
         "\n");
  printf("  \"-s\" AND/OR-search karma DB according to a \"searchString\";\n"
	 "       \"searchString\" syntax: <property>=<searchStr>[{,|+}...], "
                                                              "where: \n"
	 "       - <searchStr> is a substring of the <property> value\n"
	 "         (see ~/.openrio/*/fileinfo for possible <property> names)\n"
	 "       - \',\' and \'+\' mean AND and OR, respectively\n"
	 "       - each AND/OR operator is applied to the "
	                        "previous search results\n"
         "\n");
  printf("  \"-S\" Same as \"-s\" but <searchStr> must be an exact match\n");
  printf("  \"-u\" use properties to build download path for <type>=\'tune\'. "
	                                            "Default=\'%%f.%%c\'\n"
         "  \"-x\" like -u for <type>=\'taxi\',.    Default=\"-u props\"\n"
         "  \"-l\" like -u for <type>=\'playlist\'. Default=\"-x props\"\n");
  printf("       \"props\" are formating strings (as in \"printf\"), where\n"
         "       \'%%T\', \'%%g\', \'%%a\', \'%%s\', \'%%f\', \'%%t\', "
                                  "\'%%n\', \'%%c\' are, resp., values of\n"
	 "       <type>, <genre>, <artist>, <source>, <fid>, <title>, "
                                                 "<tracknr>, <codec>\n"
	 "\n");
  printf("  \"-e\" edit the download pathname using \"set1\" and \"set2\" "
                                                     "as in unix \'tr\'\n"
         "       - if \"set3\" is empty (\"\"): just translate as in \'tr\';\n"
	 "       - otherwise, further delete all chars given in \"set3\"\n"
	 "  \"-q\" squeeze repeated chars (given in \"set2\"),"
                                         " after the optional delete\n\n");
  printf("  \"-d\" download the files searched for with \"-s\"\n"
         "  \"-r\" remove these files (after the optional download)\n"
         "  \"-R\" same as -r, but do not ask for confirmation\n");
}

void currentKarma(int i, int *k) 
{
	static int karma;
	if (i)
		karma=*k;
	else
		*k=karma;
}

void fidNamesFileHandler(const int op, const int fid, const char * data) {
  int karma, i=0;
  uint32_t * fdb;
  char * filename;

/*fprintf(stderr, "-->%d, %d, %s\n", op, fid, data); */
  if (op==0) {                                                       /* Close */
  	lk_fdb_save();
  }
  else if ((op==1) || (op==2)) {                        /* Read or Write Open */
    if(lk_fdb_load() == -1)
	    fprintf(stderr, "** Error: can't load the fdb\n");
  }
  else if (op==3) {                    /* Download files as given in fidNames */
	  currentKarma(0, &karma);
	  fdb=lk_properties_andOrSearch(EXACT|ORS, NULL, "path", "");
	  if(fdb){
		  while(fdb[i]){
			filename=lk_properties_get_property(fdb[i], "path");
		  	printf("Downloading fid %i to %s\n", fdb[i], filename);
			fprintf(stderr, "rio_read %s\n", 
				CHECK(rio_read(karma, fdb[i], filename)));
			++i;
		  }
		  free(fdb);
	  }
  }
  else if (op==4) {            /* Write fid and filename=data to fidNamesFile */
	  lk_properties_set_property(fid, "path", (char *)data);
  }
  else if (op==5) {}                        /* Upload   fidNamesFile to   Rio */
  else if (op==6) {}                        /* Download fidNamesFile from Rio */
  else fprintf(stderr, "** Error: unknown op in fidNamesFileHandler\n");
}

int treecopy(const char *file, const struct stat *sb, int flag)
{
	static int first = 1;
	static int karma;
	uint32_t fid;

	if(first){   
		currentKarma(0, &karma);
		first=0;
	}
	if(flag==FTW_F) {
		fid = rio_write(karma, file);
		printf("Uploading fid %d, %s: %s\n", 
                                  fid, file, (fid>0)?"ok":"failed");
 		fidNamesFileHandler(4, fid, file);
	}
	return 0;
}

uint32_t * searchStr(int exact, char *searchString)
{
  int sizec=0, andOr=-1;                           /* andOr = 1 (and), 0 (or) */
  uint32_t *vc=NULL, *vp=NULL;                    /* vc/vp : current/previous */
  char *keyValue=NULL, *key=NULL;
  char *opp=NULL, op='\0';          /* op: logical operator;  opp: op pointer */

  if (searchString==NULL) return NULL;
  op = ((opp = strpbrk(searchString, ",+"))==NULL) ? '\0' : opp[0];
  keyValue=strsep(&searchString, "+,");    /* searchString -> past "," or "+" */

  while (keyValue!=NULL) {
    key=utf8_to_codeset(keyValue);
    fprintf(stderr, "%9s FIDs with \"%-30s:", 
          (andOr==-1)?"Searching":(andOr?"ANDing":"ORing"),key);
    free(key);
    if(andOr==-1) andOr = ORS;
    if(exact) andOr = EXACT|andOr;

    key=strsep(&keyValue, "=");                   /* keyValue -> past the "=" */

    if (vp){
	    vc=lk_properties_andOrSearch(andOr, vp, key, keyValue); /*frees vp*/
	    vp=NULL;
    }else vc=lk_properties_andOrSearch(andOr, NULL, key, keyValue);

    sizec=0;
    if(vc != NULL) {
      while(vc[sizec] != 0) sizec++;                            /* size of vc */
      vp = malloc((sizec+1)*sizeof(uint32_t));
      memcpy(vp, vc, (sizec+1)*sizeof(uint32_t));               /* copy vp=vc */
      free(vc);
    }
    fprintf(stderr, " %4d selected\n", sizec);

    if (searchString==NULL) break;

    andOr = (op==',');
    op = ((opp = strpbrk(searchString, ",+"))==NULL) ? '\0' : opp[0];
    keyValue=strsep(&searchString, "+,");  /* searchString -> past "," or "+" */
  }
  return vp;
}

void deleteFids(uint32_t *vp) {
  int karma, c;

  currentKarma(0, &karma);
  for (c=0; vp[c]!=0; c++) 
    if (vp[c]!=256) {                               /* ignoring root playlist */
    	fprintf(stderr, "lk_karma_delete_file(fid=%d): %s\n", 
	             vp[c], CHECK(lk_karma_delete_file(karma, vp[c])));
	lk_properties_del_property(vp[c]);
    }
}

/* -------------------------------------------------------------------------- */
/* ------------------ Pathname editing: now in pathedit.c ------------------- */
/* -------------------------------------------------------------------------- */

void downloadFids(uint32_t *vp, int doDownload,
		  char* useTuneProps, char* useTaxiProps, char* usePlstProps, 
                  char* edPathStr1, char* edPathStr2, char* edPathStr3, 
		  int squeezeRep)
{
  int karma, i, j;
  char c, cc, *useProps=NULL, filename[512]="\0";
  char *type, *genre, *artist,*source, *title, *tracknr, *codec;
  uint32_t fid;

  currentKarma(0, &karma);
  for (j=0; vp[j]!=0; j++) {
        if((!doDownload) && (j==0)) 
          printf("# Fid Name\n#---- -------------------------------------\n");
        fid=vp[j];
        type=lk_properties_get_property(fid, "type");
        if(type==NULL) {
          fprintf(stderr, "**Could not get type of fid=%d\n", fid);
          return;
        }
	if(strcmp((char *)type, "playlist") == 0) useProps=strdup(usePlstProps);
	if(strcmp((char *)type, "taxi") == 0)     useProps=strdup(useTaxiProps);
	if(strcmp((char *)type, "tune") == 0)     useProps=strdup(useTuneProps);

	for (i=0; useProps[i]!='\0'; i++) {
	  cc=(useProps[i]);
	  if (cc == '%') {
	    c=useProps[++i];
	    switch(c) {
	    case '\0':
	      fprintf(stderr, "** Format error: %s\n", useProps);
	      i--;
	    break;
	    case '%':
	      strncat(filename, useProps+i, 1);
	    break;
	    case 'T':                                               /* <type> */
	      strcat(filename, type);
	    break;
	    case 'g':                                              /*< genre> */
	      genre=lk_properties_get_property(fid, "genre");
	      if(genre) strcat(filename, genre);
	    break;
	    case 'a':                                             /* <artist> */
	      artist=lk_properties_get_property(fid, "artist");
	      artist=utf8_to_codeset(artist);
	      if(artist){
		      strcat(filename, artist);
		      free(artist);
	      }
	    break;
	    case 's':                                     /* <source> (album) */
	      source=lk_properties_get_property(fid, "source");
	      source=utf8_to_codeset(source);
	      if(source){
		      strcat(filename, source);
		      free(source);
	      }
	    break;
	    case 'f':                                                /* <fid> */
	      strcat(filename, simple_itoa(fid));
	    break;
	    case 't':                                              /* <title> */
	      title=lk_properties_get_property(fid, "title");
	      title=utf8_to_codeset(title);
	      if(title){
		      strcat(filename, title);
		      free(title);
	      }
	      else{
		fprintf(stderr, "**Could not get title of fid=%d\n", fid);
		/* return; */
	      }
	    break;
	    case 'n':                                            /* <tracknr> */
	      tracknr=lk_properties_get_property(fid, "tracknr");
	      strcat(filename, tracknr);
	    break;
	    case 'c':                                   /* <codec>|"playlist" */
	      if(strcmp((char *)type, "playlist") == 0) 
		strcat(filename, type);
	      else{
		codec=lk_properties_get_property(fid, "codec");
		if(codec) strcat(filename, codec);
		else{
		  fprintf(stderr, "**Could not get codec of fid=%d\n", fid);
		  return;
		}
	      }
	    break;
	    default:
	      fprintf(stderr, "* Unknown property \'%c\' - ignored\n", c);
	    break;
	    } /* switch(c) */
	  }else{                           /* just copy all non-special chars */
	    strncat(filename, useProps+i, 1);
	  } /* if(cc == ...) */
	} /* for i=0 ...  */
	
	free(useProps);

	if(edPathStr1 && edPathStr2 && edPathStr3)   /* 'tr'-editing filename */
	  tr(filename, edPathStr1, edPathStr2, edPathStr3, squeezeRep);

        if(doDownload) {
	  fprintf(stderr, "Downloading fid %i to %s\n", fid, filename);

	  if(strcmp((char *)type, "playlist") == 0) {
	    fprintf(stderr, "rio_read_playlist %s\n",
		             CHECK(rio_read_playlist(karma, fid, filename)));
	  }else{
	    fprintf(stderr, "rio_read %s\n", 
                             CHECK(rio_read(karma, fid, filename)));
	  }
	}else{
	  printf("%5i %s\n", fid, filename);
	}
	filename[0]='\0';
	if((!doDownload) && (vp[j+1]==0)) 
	  printf("#---- -------------------------------------\n");
  }
}

double getUsedSize(int karma)
{
	/* int rio; uint32_t storage_id;  */
	uint32_t n_files, highest_file_id;
        uint64_t s_size, f_space;
	double size, freeSpace;

        n_files=s_size=f_space=highest_file_id=0;
	lk_karma_get_storage_details(karma, 0, 
			    &n_files, &s_size, &f_space, &highest_file_id);
	size = s_size; 
	freeSpace = f_space;
/*      fprintf(stderr, "%E", size - freeSpace); */
	return(size - freeSpace);
}

int main(int argc, char * argv[])
{
	int i, c, r;
        char *settings;
	char *host=NULL;
	char *usercodeset=NULL;/*="ISO8859-15";*/
	char *psw=NULL;    /* "psw" is a bit better to search than old "p" ;) */
	char *searchString=NULL;
	char *tunePropsToUse="%f.%c"; /* props string for tune download paths */
	char *taxiPropsToUse=NULL;    /* props string for taxi download paths */
	char *plstPropsToUse=NULL;    /* props string for playlst downld paths*/
	char *editPathStr1=NULL;
	char *editPathStr2=NULL;
	char *editPathStr3=NULL;
	int  squeezeRepeats=0;
	int  download=0;
	int  fidNamesDownload=0;
	int  removeFiles=0;
	int  exact=0;
	char yesNot='N';
	int  karma;
	uint16_t port;
        uint32_t *vp=NULL;                                  /* vector of fids */
	int dbChanged;
	char *utf;
	
	while ((c = getopt(argc, argv, "dhfqrRa:c:e:p:s:S:u:x:l:")) != -1) {
		switch(c) {
		case 'h': case '?':
			usage();
			return -1;
			break;
		case 'd':
		        download=1;
			break;
		case 'f':
		        fidNamesDownload=1;
			break;
		case 'r':
		        removeFiles=1;
			break;
		case 'R':
		        removeFiles=2;    /* don't ask confirmation questions */
			break;
		case 'a':
			host=strdup(optarg);
			break;
		case 'c':
			usercodeset=strdup(optarg);
			break;
		case 'e':
		        editPathStr1=strdup(optarg); 
			if (optind>argc-2) 
			  {usage(); return -1;}
			editPathStr2=strdup(argv[optind]); optind++; 
			editPathStr3=strdup(argv[optind]); optind++;
			break;
		case 'q':
		        squeezeRepeats=1;
			break;
		case 'p':
			psw=strdup(optarg);
			break;
		case 's':
			searchString=strdup(optarg);
			break;
		case 'S':
		        searchString=strdup(optarg); exact = 1;
			break;
		case 'u':
			tunePropsToUse=strdup(optarg);
			break;
		case 'x':
			taxiPropsToUse=strdup(optarg);
			break;
		case 'l':
			plstPropsToUse=strdup(optarg);
			break;
		default:
			fprintf(stderr, "unknown option %c\n", c);
			break;
		}
	}
	if(taxiPropsToUse==NULL) taxiPropsToUse=strdup(tunePropsToUse);
	if(plstPropsToUse==NULL) plstPropsToUse=strdup(taxiPropsToUse);

	lk_errors_autoprint(0);
	if(!host) {
		for(i=0; (i<=ssdpRetry)&&(host==NULL); i++) {
			lk_errors_number();
			lk_ssdp_discover(&host, &port);
			host?fprintf(stderr, "SSDP: %s:%i\n", host, port)
			    :fprintf(stderr, "SSDP: no answer\n");
		}
	}
	if (host==NULL) {
		lk_errors_p("*", " -- Can't connect to Rio Karma");
		usage();
		if(psw) free(psw);
		if(searchString) free(searchString);
		if(usercodeset) free(usercodeset);
		return -1;
	}
	if (host!=NULL) {
	  karma=lk_karma_connect(host);
	  lk_errors_p("","");
	  free(host);
	}

	if(karma < 0) {
		usage();
		if(psw) free(psw);
		if(searchString) free(searchString);
		if(usercodeset) free(usercodeset);
		return -1;
	}
	
	currentKarma(1,&karma);
	
	/*lk_properties_init("ISO8859-15");*/
	if(usercodeset != NULL){
		lk_properties_init();
		utf8_set_codeset(usercodeset);
		if (lk_errors_p("", "; Trying locale...")) {
			free(usercodeset);
			usercodeset = NULL;
		}
	}
	if(usercodeset == NULL){
                if (! setlocale(LC_CTYPE, ""))
                  fprintf(stderr,"* Warning: setlocale failed.");
		if(strlen(usercodeset = nl_langinfo(CODESET)) == 0) {
			fprintf(stderr, "* Warning: nl_langinfo problem; "
				"using default codeset %s\n", defaultCodeset);
			usercodeset = defaultCodeset;
		}
		lk_properties_init();
		utf8_set_codeset(usercodeset);
 		if(! lk_errors_p("",""))
		  fprintf(stderr, "* Using %s codeset.\n", usercodeset);
		/*free(usercodeset);*/
	}
	fprintf(stderr, "Using %s codeset.\n", usercodeset);
	lk_errors_autoprint(1);               /* Setting auto-print-errors ON */
	
	if(psw){
		fprintf(stderr, "lk_karma_authenticate: %i\n", 
			         lk_karma_authenticate(karma, psw));
		free(psw);
	}else   fprintf(stderr, "lk_karma_authenticate: %i\n", 
			         lk_karma_authenticate(karma, ""));
	r = lk_karma_request_io_lock(karma,IO_LOCK_R);
	fprintf(stderr, "lk_karma_request_io_lock: %s\n", CHECK(r));
	if(r == -1){
		printf("Access Denied!\n");
		lk_karma_hangup(karma);
		close(karma);
		lk_properties_destroy();
		utf8_destroy();
		return -1;
	}

        lk_karma_get_device_settings(karma, &settings);
	dbChanged = lk_properties_cache_obsolete(settings);
        if(dbChanged) {
	  printf("Downloading Rio database... ");
	  fflush(stdout);
	  if(getUsedSize(karma)>3E9) printf(" please wait... ");
	  fflush(stdout);
	}
	free(settings);
	if(!r) lk_karma_load_database(karma);
       	if(dbChanged) printf("Done!\n");

        if(fidNamesDownload) {
	  fidNamesFileHandler(1, 0, NULL);
	  fidNamesFileHandler(3, 0, NULL);
	}

	if (searchString != NULL){
		utf=utf8_from_codeset(searchString);
		free(searchString);
		vp=searchStr(exact, utf);
		free(utf);
	}
	if(vp) downloadFids(vp, download, 
		     tunePropsToUse, taxiPropsToUse, plstPropsToUse, 
		     editPathStr1, editPathStr2, editPathStr3, squeezeRepeats);
	
	if(taxiPropsToUse) free(taxiPropsToUse); /*Enrique?*/
	if(plstPropsToUse) free(plstPropsToUse); /*Enrique?*/

	if(removeFiles>0 && vp) {
	  yesNot='Y';  for (i=0; vp[i]!=0; i++) continue;
	  if(i>0) {
	    printf("Ready to remove %d files; ", i);
	    if(removeFiles==1) {
	      printf("proceed (Y/N)? ");  scanf("%c", &yesNot);
	    }else{
	      printf("\n");
	    }
	    if(yesNot!='Y' && yesNot!='y') 
	      printf("Remove aborted!\n");
	    else {
	      fprintf(stderr, "lk_karma_request_io_lock (W): %s\n", 
		              CHECK(lk_karma_request_io_lock(karma,IO_LOCK_W)));
	      deleteFids(vp);
	    }
	  }
	  else fprintf(stderr, "No files to remove!\n");
	}

	if(optind < argc){
		fprintf(stderr, "lk_karma_request_io_lock: %s\n", 
			      CHECK(lk_karma_request_io_lock(karma,IO_LOCK_W)));
		if(synchronize_necessary(karma)){
			lk_karma_get_all_file_details(karma, &psw);
			lk_properties_import(psw);
			free(psw);
		}
		fidNamesFileHandler(2, 0, NULL);
		for(; optind<argc; optind++)
			ftw(argv[optind], treecopy, 1);
	}
	fprintf(stderr, "lk_karma_release_io_lock: %s\n", 
		   CHECK(lk_karma_release_io_lock(karma)));

	fprintf(stderr, "lk_karma_hangup: %s\n", CHECK(lk_karma_hangup(karma)));

	free(vp);
	close(karma);
	fidNamesFileHandler(0, 0, NULL);
	lk_properties_save();
	lk_properties_destroy();
	utf8_destroy();
	return 0;
}
